// vim:ts=4:sts=4:sw=4:noet:tw=72

package ircd

import (
	"fmt"
	"net"
	"os"
	"runtime"
	"strconv"
	"strings"
	"time"

	"code.dnix.de/an/conf"
	"code.dnix.de/an/irc"
	"code.dnix.de/an/xlog"
)

const (
	CHANLIMIT  = 1024
	CHANNELLEN = 200
	TOPICLEN   = 1024
)

var myinfo string = "%s %s/%s * *"
var isupport string = "ALIAS FRIEND UNFRIEND CASEMAPPING=rfc1459 CHANLIMIT=#:1024 CHANMODES=b,k,l,imnpst CHANNELLEN=200 CHANTYPES=# EXCEPTS=e KICKLEN MAXLIST=b:50,e:50 MODES=1 NETWORK=dnix.de NICKLEN=32 PREFIX=(aohv)&@%%+ SAFELIST STATUSMSG=&@%%+ TOPICLEN"

type Server struct {
	Dispatch  chan *irc.Message
	AddClient chan Client
	DelClient chan Client

	host       string
	info       string
	software   string
	version    string
	created    string
	motd       string
	clients    map[string]Client
	channels   map[string]map[string]string
	ports      map[int]bool
	config     *conf.ConfigFile
	configPath string
}

func init() {
}

// Create a new server instance.
func NewServer(configPath, software, version string) *Server {
	srv := &Server{software: software, version: version, created: "yes"}

	srv.Dispatch = make(chan *irc.Message, 1024)
	srv.AddClient = make(chan Client, 1024)
	srv.DelClient = make(chan Client, 1024)

	srv.clients = make(map[string]Client)
	srv.channels = make(map[string]map[string]string)
	srv.configPath = configPath
	srv.loadConfig()
	loglevel, _ := srv.config.GetInt("system", "loglevel")
	srv.host, _ = srv.config.GetString("server", "host")
	srv.info, _ = srv.config.GetString("server", "info")
	srv.motd, _ = srv.config.GetString("server", "motd")
	xlog.Init(loglevel)

	return srv
}

// Open the listening port and start the main server loop.
func (srv *Server) Run() {
	xlog.Info("%s/%s", srv.software, srv.version)
	address, _ := srv.config.GetString("net", "address")
	port, _ := srv.config.GetInt("net", "port")
	go srv.listen(address, port)
	srv.dispatch()
}

func (srv *Server) listen(address string, port int) {
	if _, exists := srv.ports[port]; exists {
		xlog.Warning("Port %i already opened", port)
	}
	listen, err := net.Listen("tcp", address+":"+strconv.Itoa(port))
	if err != nil {
		xlog.Fatal("Cannot listen on port %i (%s), exiting",
			port, err.Error())
		os.Exit(-1)
	}

	xlog.Info("Start listening on port %d", port)

	for {
		time.Sleep(1e6)
		conn, err := listen.Accept()
		if err != nil {
			xlog.Error(err.Error())
		} else {
			NewRemoteClient(srv, conn)
		}
	}
}

func (srv *Server) dispatch() {
	xlog.Debug("Entering msg dispatcher")
	for {
		time.Sleep(1 * time.Millisecond)
		select {
		case msg := <-srv.Dispatch:
			srv.recvMsg(msg)
		case cl := <-srv.AddClient:
			name := cl.Name()
			if _, exists := srv.clients[name]; exists {
				xlog.Warning("Client registration failed: '%s'", name)
				go func() { cl.Register() <- false }()
				continue
			}
			go func() { cl.Register() <- true }()
			srv.clients[name] = cl
			xlog.Info("Client registered: '%s'", name)
			xlog.Info("Server has %d client(s)", len(srv.clients))
			xlog.Debug("Goroutines running: %d", runtime.NumGoroutine())
			srv.clientLogon(cl)
			srv.clientMotd(cl)
		case cl := <-srv.DelClient:
			name := cl.Name()
			cl.Destroy()
			//go func() {
			//	time.Sleep(10 * time.Second)
			delete(srv.clients, name)
			//}()
			xlog.Info("Client deleted: '%s'", name)
			xlog.Info("Server has %d client(s)", len(srv.clients))
			xlog.Debug("Goroutines running: %d", runtime.NumGoroutine())
		default:
		}
	}
}

func (srv *Server) loadConfig() {
	cfg, err := conf.ReadConfigFile(srv.configPath)
	if err != nil {
		xlog.Fatal("Can't read config file (%s)", err.Error())
		os.Exit(-1)
	}
	srv.config = cfg
}

func (srv *Server) recvMsg(msg *irc.Message) {
	cmd := msg.Cmd
	hook, exists := srvCommandHooks[cmd]
	if !exists {
		srv.sendReply(msg.Pre, ERR_UNKNOWNCOMMAND, cmd, "Unknown command")
		return
	}
	argc := len(msg.Args)
	if argc < hook.MinArgs {
		srv.sendReply(msg.Pre, ERR_NEEDMOREPARAMS, cmd, "Not enough parameters")
		return
	}
	hook.HookFn(srv, msg)
}

func (srv *Server) sendMsg(msg *irc.Message) {
	if strings.HasPrefix(msg.Args[0], "#") {
		srv.channelBroadcast(msg.Args[0], msg)
	} else {
		srv.sendMsgToClient(msg)
	}
}

func (srv *Server) sendMsgToChannel(msg *irc.Message) {
	cname := msg.Args[0]
	if _, exists := srv.channels[cname]; !exists {
		return
	}
	subs := srv.channels[cname]
	for sub, _ := range subs {
		if _, exists := srv.clients[sub]; !exists {
			continue
		}
		cl := srv.clients[sub]
		cl.Receive(msg)
	}
}

func (srv *Server) sendMsgToClient(msg *irc.Message) {
	if _, exists := srv.clients[msg.Args[0]]; !exists {
		xlog.Error("sendMsgToClient: Client '%s' does not exist", msg.Args[0])
		return
	}
	cl := srv.clients[msg.Args[0]]
	cl.Receive(msg)
}

func (srv *Server) sendReply(tar, cmd, args, trail string) {
	if _, exists := srv.clients[tar]; !exists {
		xlog.Error("sendReply: Client '%s' does not exist", tar)
		return
	}
	cl := srv.clients[tar]
	if args != "" {
		args = tar + args
	} else {
		args = tar
	}
	cl.Receive(irc.M(srv.host, cmd, args, trail))
}

func (srv *Server) clientLogon(cl Client) {
	srv.sendReply(cl.Name(), RPL_WELCOME, "", "Willkommen!")
	srv.sendReply(cl.Name(), RPL_YOURHOST, "",
		fmt.Sprintf("Your host is %s, running on %s/%s",
			srv.host, srv.software, srv.version))
	srv.sendReply(cl.Name(), RPL_CREATED, "",
		fmt.Sprintf("Created: %s", srv.created))
	srv.sendReply(cl.Name(), RPL_MYINFO, "",
		fmt.Sprintf(myinfo, srv.host, srv.software, srv.version))
	srv.sendReply(cl.Name(), RPL_ISUPPORT, "",
		isupport+" are supported by this server")
}

func (srv *Server) clientMotd(cl Client) {
	srv.sendReply(cl.Name(), RPL_MOTDSTART, "",
		fmt.Sprintf("- %s Message of the day -", srv.host))
	for _, line := range strings.Split(srv.motd, "\n") {
		srv.sendReply(cl.Name(), RPL_MOTD, "", fmt.Sprintf("- %s", line))
	}
	srv.sendReply(cl.Name(), RPL_ENDOFMOTD, "", "End of MOTD command")
}

func (srv *Server) channelBroadcast(ch string, msg *irc.Message) {
	for nick, _ := range srv.channels[ch] {
		if msg.Pre == nick && msg.Cmd == "PRIVMSG" {
			continue
		}
		srv.clients[nick].Receive(msg)
	}
}

func (srv *Server) channelJoin(nick, ch string) {
	if _, exists := srv.channels[ch]; !exists {
		srv.channels[ch] = make(map[string]string)
	}
	if _, exists := srv.channels[ch][nick]; exists {
		return
	}
	srv.channels[ch][nick] = ""
	srv.channelBroadcast(ch, irc.M(nick, "JOIN", ch, ""))
}

func (srv *Server) channelPart(nick, ch, reason string) {
	if _, exists := srv.channels[ch]; !exists {
		return
	}
	if _, exists := srv.channels[ch][nick]; !exists {
		return
	}
	srv.channelBroadcast(ch, irc.M(nick, "PART", ch, reason))
	delete(srv.channels[ch], nick)
}

type SrvCommandHook struct {
	HookFn   func(srv *Server, msg *irc.Message)
	MinArgs  int
	NeedOper bool
	NeedAuth bool
}

var srvCommandHooks = map[string]SrvCommandHook{
	"PRIVMSG":  {srvHandleCmdPrivmsg, 1, false, false},
	"JOIN":     {srvHandleCmdJoin, 1, false, false},
	"PART":     {srvHandleCmdPart, 1, false, false},
	"QUIT":     {srvHandleCmdQuit, 0, false, false},
	"MODE":     {srvHandleCmdMode, 1, false, false},
	"LIST":     {srvHandleCmdList, 0, false, false},
	"VERSION":  {srvHandleCmdVersion, 0, false, false},
	"STATS":    {srvHandleCmdStats, 0, false, false},
	"TIME":     {srvHandleCmdTime, 0, false, false},
	"OPER":     {srvHandleCmdOper, 1, false, false},
	"ADMIN":    {srvHandleCmdAdmin, 0, false, false},
	"INFO":     {srvHandleCmdInfo, 0, false, false},
	"WHO":      {srvHandleCmdWho, 0, false, false},
	"WHOIS":    {srvHandleCmdWhois, 0, false, false},
	"WHOWAS":   {srvHandleCmdWhowas, 0, false, false},
	"KILL":     {srvHandleCmdKill, 0, false, false},
	"PING":     {srvHandleCmdPing, 0, false, false},
	"PONG":     {srvHandleCmdPong, 0, false, false},
	"ERROR":    {srvHandleCmdError, 0, false, false},
	"AWAY":     {srvHandleCmdAway, 0, false, false},
	"REHASH":   {srvHandleCmdRehash, 0, false, false},
	"RESTART":  {srvHandleCmdRestart, 0, false, false},
	"SUMMON":   {srvHandleCmdSummon, 0, false, false},
	"USERS":    {srvHandleCmdUsers, 0, false, false},
	"USERHOST": {srvHandleCmdUserhost, 0, false, false},
	"ISON":     {srvHandleCmdIson, 0, false, false},
}

func srvHandleCmdPrivmsg(srv *Server, msg *irc.Message) {
	srv.sendMsg(msg)
}

func srvHandleCmdJoin(srv *Server, msg *irc.Message) {
	srv.channelJoin(msg.Pre, msg.Args[0])
}

func srvHandleCmdPart(srv *Server, msg *irc.Message) {
	srv.channelPart(msg.Pre, msg.Args[0], msg.Trail)
}

func srvHandleCmdOper(srv *Server, msg *irc.Message) {
}

func srvHandleCmdQuit(srv *Server, msg *irc.Message) {
}

func srvHandleCmdMode(srv *Server, msg *irc.Message) {
}

func srvHandleCmdList(srv *Server, msg *irc.Message) {
}

func srvHandleCmdVersion(srv *Server, msg *irc.Message) {
}

func srvHandleCmdStats(srv *Server, msg *irc.Message) {
}

func srvHandleCmdTime(srv *Server, msg *irc.Message) {
}

func srvHandleCmdAdmin(srv *Server, msg *irc.Message) {
}

func srvHandleCmdInfo(srv *Server, msg *irc.Message) {
}

func srvHandleCmdWho(srv *Server, msg *irc.Message) {
}

func srvHandleCmdWhois(srv *Server, msg *irc.Message) {
	srv.sendReply(msg.Pre, RPL_WHOISUSER, "nick user host *", "real name")
}

func srvHandleCmdWhowas(srv *Server, msg *irc.Message) {
}

func srvHandleCmdKill(srv *Server, msg *irc.Message) {
}

func srvHandleCmdPing(srv *Server, msg *irc.Message) {
	srv.sendReply(msg.Pre, "PONG", msg.Args[0], "")
}

func srvHandleCmdPong(srv *Server, msg *irc.Message) {
}

func srvHandleCmdError(srv *Server, msg *irc.Message) {
}

func srvHandleCmdAway(srv *Server, msg *irc.Message) {
}

func srvHandleCmdRehash(srv *Server, msg *irc.Message) {
	srv.loadConfig()
	srv.sendReply(msg.Pre, RPL_REHASHING, "", "Rehashing.")
	xlog.Info("Rehashing")
}

func srvHandleCmdRestart(srv *Server, msg *irc.Message) {
}

func srvHandleCmdSummon(srv *Server, msg *irc.Message) {
}

func srvHandleCmdUsers(srv *Server, msg *irc.Message) {
}

func srvHandleCmdUserhost(srv *Server, msg *irc.Message) {
}

func srvHandleCmdIson(srv *Server, msg *irc.Message) {
}
